// TODO: Put public facing types in this file.

import 'dart:convert';
import 'dart:core';
import 'dart:io';

import 'package:xdatabase/src/connection_pool.dart';

import 'table_meta.dart';

abstract class Bean<M> {
  Map<String, dynamic> asMap();
  M fromMap(Map<String, dynamic> map);
  Future<int> insert(QueryAble conn);
  Future<int> update(String statement, QueryAble conn);
  Future<int> delete(String statement, QueryAble conn);
  Future<int> updateByPK(QueryAble conn);
  Future<int> deleteByPk(QueryAble conn);
  Future<M?> selectByPk(QueryAble conn);
  Future<List<M>> select(String statement, QueryAble conn);
}

String symbolToName(Symbol symbol) {
  final bs = symbol.toString();
  return bs.substring(8, bs.length - 2);
}

String nameUnderLine(String original) {
  return original
      .split('')
      .map((w) {
        if (w.toLowerCase() == w) {
          return w;
        } else {
          return '_${w.toLowerCase()}';
        }
      })
      .toList()
      .join('');
}

class Table {
  final String tableName;
  const Table(this.tableName);
}

class Index {
  final String name;
  final List<String> columns;
  const Index({required this.name, required this.columns});
}

class UniqueIndex {
  final String name;
  final List<String> columns;
  const UniqueIndex({required this.name, required this.columns});
}

class Field {
  final String type;
  final bool primary;
  final String mark;
  const Field({required this.type, this.primary = false, this.mark = ''});
}

class DeclarationHolder {
  Declarations declarations;
  Map<Type, TableDeclaration> tables;
  Map<String, TableDeclaration> queryTables;
  DeclarationHolder._privateConstructor()
      : declarations = Declarations(),
        tables = <Type, TableDeclaration>{},
        queryTables = <String, TableDeclaration>{};
  static final DeclarationHolder _instance =
      DeclarationHolder._privateConstructor();
  factory DeclarationHolder() {
    return _instance;
  }

  TableDeclaration? ofTable<T>() {
    if (tables.containsKey(T)) {
      return tables[T]!;
    }
    return null;
  }

  TableDeclaration? byTable(String className) {
    if (queryTables.containsKey(className)) {
      return queryTables[className];
    }
  }

  void saveToPath(String path) {
    final map = declarations.asMap();
    File(path).writeAsStringSync(jsonEncode(map));
  }

  void fromPersists(String persists) {
    final map = jsonDecode(persists);
    declarations = Declarations.fromMap(map);
    declarations.tables.forEach((tableName, tableDeclaration) {
      queryTables[tableDeclaration.className] = tableDeclaration;
    });
  }

  String compareTo(Declarations other) {
    final tableNeedsToCreate = declarations.tables.keys
        .where((e) => !other.tables.containsKey(e))
        .toList();
    final tableNeedsToDelete = other.tables.keys
        .where((element) => !declarations.tables.containsKey(element))
        .toList();
    var lines = [];
    lines.addAll(tableNeedsToCreate
        .map((tableName) => declarations.tables[tableName]!.createScript())
        .toList());
    lines.addAll(tableNeedsToDelete
        .map((tableName) => other.tables[tableName]!.dropTableSql()));
    for (var tableDeclaration in declarations.tables.values) {
      for (var currentDeclaration in other.tables.values) {
        if (tableDeclaration.tableName == currentDeclaration.tableName) {
          lines.add(tableDeclaration.compareWith(currentDeclaration));
        }
      }
    }
    return lines.join('\n\n');
  }

  String compareWithPath(String path) {
    final declarationFile = File(path);
    if (!declarationFile.existsSync()) {
      return declarations.createTables();
    }
    final jtext = declarationFile.readAsStringSync();
    final defineMap = jsonDecode(jtext);
    final currentDeclaration = Declarations.fromMap(defineMap);
    return compareTo(currentDeclaration);
  }
}

class AndCondition {
  late List<Column> columns;
  AndCondition() : columns = <Column>[];
  AndCondition.init(Column first, Column second) {
    columns = <Column>[];
    columns.add(first);
    columns.add(second);
  }

  @override
  String toString() {
    columns.map((c) => c.toString()).toList().join(' AND ');
    return super.toString();
  }

  AndCondition and(Column col) {
    columns.add(col);
    return this;
  }
}

class Column {
  String fieldName;
  String? operator;
  Column(this.fieldName);
  String? order;
  Column eq() {
    operator = '=';
    return this;
  }

  Column lt() {
    operator = '<';
    return this;
  }

  Column gt() {
    operator = '>';
    return this;
  }

  Column lte() {
    operator = '<=';
    return this;
  }

  Column gte() {
    operator = '>=';
    return this;
  }

  Column like() {
    operator = ' like ';
    return this;
  }

  AndCondition and(Column right) {
    return AndCondition.init(this, right);
  }

  Column desc() {
    order = '$fieldName DESC';
    return this;
  }

  Column asc() {
    order = '$fieldName ASC';
    return this;
  }

  String orderField() {
    return order!;
  }

  @override
  String toString() => '$fieldName $operator @$fieldName';
}
